# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""This package contains utilities for PyTorch verifying datasets."""

from openfl.federated.data.sources.torch.verifiable_map_style_dataset import (
    VerifiableMapStyleDataset,
)
from openfl.federated.data.sources.torch.verifiable_map_style_image_folder import (
    VerifiableImageFolder,
)
